#ifndef WIN9X

#include "Console.h"
#include "LoadGameHook.h"
#include "SafeWrite.h"

static bool active=false;

static ID3DXFont* font;
static ID3DXSprite* sprite;

#define MAXLINES (12)

static DWORD firstline;
static DWORD linecount;
static DWORD nextline;
static DWORD linewidth;
static char* lines[MAXLINES];
static char* line;
static DWORD showingline;
static DWORD visiblelines=5;

static RECT* rects;

static void scroll(int amount) {
	if(amount==0||linecount<visiblelines) return;
	if(amount>0) {
		for(int i=0;i<amount;i++) {
			if(showingline+visiblelines==nextline) return;
			showingline=(showingline+1)%MAXLINES;
		}
	} else {
		for(int i=0;i<amount;i++) {
			if(showingline==firstline) return;
			showingline--;
			if(showingline==-1) showingline=MAXLINES-1;
		}
	}
}
static void AddLine(const char* msg) {
	strcpy(lines[nextline++], msg);
	if(linecount==MAXLINES) firstline++;
	else linecount++;
	if(nextline==MAXLINES) nextline=0;
	if(firstline==MAXLINES) firstline=0;

	if(linecount>visiblelines) {
		showingline=nextline-visiblelines;
		if(nextline<visiblelines) showingline+=MAXLINES;
	}
}
static void _stdcall ConsoleHook2(const char* msg) {
	DWORD pos=0;
	DWORD lastspace=-1;
	for(DWORD i=0;i<strlen(msg);i++) {
		if(msg[i]=='\r') continue;
		if(msg[i]=='\n') {
			line[pos]=0;
			AddLine(line);
			pos=0;
			lastspace=-1;
		} else {
			if(msg[i]==' ') lastspace=pos;
			line[pos++]=msg[i];
			if(pos==linewidth+1) {
				if(msg[i+1]==0||msg[i+1]==' ') {
					line[pos]=0;
				} else if(lastspace!=-1) {
					line[lastspace]=0;
					i-=(pos-lastspace);
				} else {
					line[pos]=0;
				}
				AddLine(line);
				pos=0;
				lastspace=-1;
			}
		}
	}
	if(pos>0) {
		line[pos]=0;
		AddLine(line);
	}
}

static void __declspec(naked) ConsoleHook() {
	_asm {
		pushad;
		push eax;
		call ConsoleHook2;
		popad;
		xor eax, eax;
		retn;
	}
}
void ConsoleClear() {
	if(!active) return;

	firstline=0;
	linecount=0;
	nextline=0;
	showingline=0;
}
void ConsoleRedraw() {
	if(!active||linecount==0) return;
	if(GetCurrentLoops()&0x7b5) return;
	sprite->Begin(D3DXSPRITE_ALPHABLEND);
	for(DWORD i=0;i<min(visiblelines, linecount);i++) {
		font->DrawTextA(sprite, lines[(showingline+i)%MAXLINES], -1, &rects[i], DT_SINGLELINE|DT_NOCLIP, 0xff00ff00);
	}
	sprite->End();
}

void ConsoleResetDevice() {
	if(!active) return;
	sprite->OnResetDevice();
	font->OnResetDevice();
}

void ConsoleLostDevice() {
	if(!active) return;
	sprite->OnLostDevice();
	font->OnLostDevice();
}

void ConsoleInit(IDirect3DDevice9* device, DWORD width, DWORD height) {
	//SafeWrite8(0x00431A78, 0xc3);
	//return;
	active=true;

	D3DXCreateFontA(device, 14, 7, 400, 0, 0, DEFAULT_CHARSET, 0, DEFAULT_QUALITY, 0, "courier new", &font);
	D3DXCreateSprite(device, &sprite);

	D3DXFONT_DESCA desc;
	font->GetDescA(&desc);
	desc.Width+=1;

	//26, 405
	//162, 56

	DWORD left=(DWORD)((26.0f*(float)width)/640.0f);
	DWORD top=(DWORD)((405.0f*(float)height)/480.0f);
	DWORD rwidth=(DWORD)((162.0f*(float)width)/640.0f);
	DWORD rheight=(DWORD)((56.0f*(float)height)/480.0f);

	visiblelines=(rheight-rheight%desc.Height)/desc.Height;
	linewidth=(rwidth-rwidth%desc.Width)/desc.Width;

	for(DWORD i=0;i<MAXLINES;i++) {
		lines[i]=new char[linewidth+1];
		memset(lines[i], 0, linewidth+1);
	}
	line=new char[linewidth];

	ConsoleClear();

	SafeWrite8(0x0043186C, 0xe9);
	SafeWrite32(0x0043186D, (DWORD)ConsoleHook - 0x00431871);

	

	rects=new RECT[visiblelines];
	for(DWORD i=0;i<visiblelines;i++) {
		rects[i].left=left;
		rects[i].right=left+width;
		rects[i].top=top+(i*desc.Height);
		rects[i].bottom=top+(i+1)*desc.Height;
	}
}

#endif