#define STRICT
#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include "dinput.h"
#include "queue9x.cpp"
#ifndef WIN9X
#include <math.h>
#include "graphics.h"
#endif
#include "ScriptExtender.h"
#include "main.h"

typedef HRESULT (_stdcall *DInputCreateProc)(HINSTANCE a,DWORD b,IDirectInputA** c,IUnknown* d);

static bool UseScrollWheel;
static DWORD WheelMod;

static bool MiddleMouseDown;
static DWORD MiddleMouseKey;

static bool BackgroundKeyboard;
static bool BackgroundMouse;

static bool AdjustMouseSpeed;
static double MouseSpeedMod;
static double MousePartX;
static double MousePartY;

static DWORD KeysDown[256];

#ifndef WIN9X
static int mouseX;
static int mouseY;

static DWORD ForcingGraphicsRefresh=0;

void _stdcall ForceGraphicsRefresh(DWORD d) { ForcingGraphicsRefresh=(d==0)?0:1; }
void GetMouse(int* x, int* y) { *x=mouseX; *y=mouseY; mouseX=0; mouseY=0; }
#endif

#define kDeviceType_KEYBOARD 0
#define kDeviceType_MOUSE 1

static queue<DIDEVICEOBJECTDATA> bufferedPresses;

DWORD _stdcall KeyDown(BYTE key) {
	return KeysDown[key];
}

void _stdcall TapKey(DWORD key) {
	DIDEVICEOBJECTDATA data;
	data.dwTimeStamp=0;
	data.dwSequence=0;
	data.dwOfs=key;
	data.dwData=0x80;
	bufferedPresses.push(data);
	data.dwData=0x00;
	bufferedPresses.push(data);
}

class FakeDirectInputDevice : public IDirectInputDeviceA {
private:
    IDirectInputDeviceA* RealDevice;
    DWORD DeviceType;
	ULONG Refs;
public:
    /*** Constructor and misc functions ***/
    FakeDirectInputDevice(IDirectInputDevice* device,DWORD type) {
        RealDevice=device;
        DeviceType=type;
		Refs=1;
    }
    /*** IUnknown methods ***/
    HRESULT _stdcall QueryInterface (REFIID riid, LPVOID * ppvObj) { return RealDevice->QueryInterface(riid,ppvObj); }
    ULONG _stdcall AddRef(void) { return ++Refs; }
    ULONG _stdcall Release(void) {
		if(--Refs==0) {
			RealDevice->Release();
			delete this;
			return 0;
		} else { return Refs; }
	}

    /*** IDirectInputDevice8A methods ***/
    HRESULT _stdcall GetCapabilities(LPDIDEVCAPS a) { return RealDevice->GetCapabilities(a); }
    HRESULT _stdcall EnumObjects(LPDIENUMDEVICEOBJECTSCALLBACKA a,LPVOID b,DWORD c) { return RealDevice->EnumObjects(a,b,c); }
    HRESULT _stdcall GetProperty(REFGUID a,DIPROPHEADER* b) { return RealDevice->GetProperty(a,b); }
    HRESULT _stdcall SetProperty(REFGUID a,const DIPROPHEADER* b) { return RealDevice->SetProperty(a,b); }
    HRESULT _stdcall Acquire(void) { return RealDevice->Acquire(); }
    HRESULT _stdcall Unacquire(void) { return RealDevice->Unacquire(); }
	//Only called for the mouse
    HRESULT _stdcall GetDeviceState(DWORD a,LPVOID b) {
#ifndef WIN9X
		if(ForcingGraphicsRefresh) RefreshGraphics();
#endif
		if(DeviceType!=kDeviceType_MOUSE) {
			return RealDevice->GetDeviceState(a,b);
		}

		RunGlobalScripts2();

        DIMOUSESTATE* MouseState=(DIMOUSESTATE*)b;
        HRESULT hr=RealDevice->GetDeviceState(sizeof(DIMOUSESTATE),MouseState);
        if(FAILED(hr)) return hr;
		if(AdjustMouseSpeed) {
			double d=((double)MouseState->lX)*MouseSpeedMod + MousePartX;
			MousePartX = modf(d, &d);
			MouseState->lX=(LONG)d;
			d=((double)MouseState->lY)*MouseSpeedMod + MousePartY;
			MousePartY = modf(d, &d);
			MouseState->lY=(LONG)d;
		}
		if(UseScrollWheel) {
			if(MouseState->lZ>0) {
				int count;
				if(WheelMod) count=MouseState->lZ/WheelMod;
				else count=1;
				while(count--) TapKey(DIK_UP);
			} else if(MouseState->lZ<0) {
				int count;
				if(WheelMod) count=(-MouseState->lZ)/WheelMod;
				else count=1;
				while(count--) TapKey(DIK_DOWN);
			}
		}
		if(MiddleMouseKey&&MouseState->rgbButtons[2]) {
			if(!MiddleMouseDown) {
				TapKey(MiddleMouseKey);
				MiddleMouseDown=true;
			}
		} else MiddleMouseDown=false;
#ifndef WIN9X
		mouseX=MouseState->lX;
		mouseY=MouseState->lY;
#endif
        return 0;
    }
	//Only called for the keyboard
	HRESULT _stdcall GetDeviceData(DWORD a,DIDEVICEOBJECTDATA* b,DWORD* c,DWORD d) {
		if(DeviceType!=kDeviceType_KEYBOARD) {
			return RealDevice->GetDeviceData(a,b,c,d);
		}
		if(!b||bufferedPresses.empty()||(d&DIGDD_PEEK)) {
			HRESULT hr=RealDevice->GetDeviceData(a,b,c,d);
			if(FAILED(hr)||!b||!(*c)) return hr;
			for(DWORD i=0;i<*c;i++) {
				KeysDown[b[i].dwOfs]=b[i].dwData&0x80;
			}
			return hr;
		}
		//Despite passing an array of 32 data objects, fallout cant seem to cope with a key being pressed and released in the same frame...
		//TODO: Fallouts behaviour when passing multiple keypresses makes it appear like it's expecting the DIDEVICEOBJECTDATA struct to be
		//      something other than 16 bytes. afaik, fallout uses DX3 for input, which is before the appData field was added, but it could
		//      be worth checking anyway.
		*b=bufferedPresses.front();
		bufferedPresses.pop();
		*c=1;
		return DI_OK;
	}
    HRESULT _stdcall SetDataFormat(const DIDATAFORMAT* a) { return RealDevice->SetDataFormat(a); }
    HRESULT _stdcall SetEventNotification(HANDLE a) { return RealDevice->SetEventNotification(a); }
    HRESULT _stdcall SetCooperativeLevel(HWND a,DWORD b) {
		if(DeviceType==kDeviceType_KEYBOARD&&BackgroundKeyboard) b=DISCL_BACKGROUND|DISCL_NONEXCLUSIVE;
		if(DeviceType==kDeviceType_MOUSE&&BackgroundMouse) b=DISCL_BACKGROUND|DISCL_NONEXCLUSIVE;
		return RealDevice->SetCooperativeLevel(a,b);
	}
    HRESULT _stdcall GetObjectInfo(LPDIDEVICEOBJECTINSTANCEA a,DWORD b,DWORD c) { return RealDevice->GetObjectInfo(a,b,c); }
    HRESULT _stdcall GetDeviceInfo(LPDIDEVICEINSTANCEA a) { return RealDevice->GetDeviceInfo(a); }
    HRESULT _stdcall RunControlPanel(HWND a,DWORD b) { return RealDevice->RunControlPanel(a,b); }
    HRESULT _stdcall Initialize(HINSTANCE a,DWORD b,REFGUID c) { return RealDevice->Initialize(a,b,c); }
};
class FakeDirectInput : public IDirectInputA {
private:
    IDirectInputA* RealInput;
	ULONG Refs;
public:
    /*** Constructor ***/
    FakeDirectInput(IDirectInput* Real) { RealInput=Real; Refs=1; }
    /*** IUnknown methods ***/
    HRESULT _stdcall QueryInterface (REFIID riid, LPVOID* ppvObj) { return RealInput->QueryInterface(riid,ppvObj); }
	ULONG _stdcall AddRef(void) { return ++Refs; }
    ULONG _stdcall Release(void) {
		if(--Refs==0) {
			RealInput->Release();
			delete this;
			return 0;
		} else { return Refs; }
    }
    /*** IDirectInput8A methods ***/
    HRESULT _stdcall CreateDevice(REFGUID r,IDirectInputDeviceA** device,IUnknown* unused) {
		GUID GUID_SysMouse    = { 0x6F1D2B60, 0xD5A0, 0x11CF, { 0xBF,0xC7,0x44,0x45,0x53,0x54,0x00,0x00} };
		GUID GUID_SysKeyboard = { 0x6F1D2B61, 0xD5A0, 0x11CF, { 0xBF,0xC7,0x44,0x45,0x53,0x54,0x00,0x00} };

        if(r!=GUID_SysKeyboard&&r!=GUID_SysMouse) { 
			return RealInput->CreateDevice(r,device,unused);
		} else {
            DWORD d;
			IDirectInputDeviceA* RealDevice;
			HRESULT hr;

            if(r==GUID_SysKeyboard) d=kDeviceType_KEYBOARD;
            else d=kDeviceType_MOUSE;
            hr=RealInput->CreateDevice(r,&RealDevice,unused);
            if(hr!=DI_OK) return hr;
            *device=new FakeDirectInputDevice(RealDevice,d);
            return DI_OK;
        }
    }
    HRESULT _stdcall EnumDevices(DWORD a,LPDIENUMDEVICESCALLBACKA b,void* c,DWORD d) { return RealInput->EnumDevices(a,b,c,d); }
    HRESULT _stdcall GetDeviceStatus(REFGUID r) { return RealInput->GetDeviceStatus(r); }
    HRESULT _stdcall RunControlPanel(HWND a,DWORD b) { return RealInput->RunControlPanel(a,b); }
    HRESULT _stdcall Initialize(HINSTANCE a,DWORD b) { return RealInput->Initialize(a,b); }
};

HRESULT _stdcall FakeDirectInputCreate(HINSTANCE a,DWORD b,IDirectInputA** c,IUnknown* d) {
	ZeroMemory(KeysDown,sizeof(KeysDown));

#ifdef WIN9X
	bufferedPresses=queue<DIDEVICEOBJECTDATA>();
#endif

	HMODULE dinput=LoadLibraryA("dinput.dll");
	if(!dinput||dinput==INVALID_HANDLE_VALUE) return -1;
	DInputCreateProc proc=(DInputCreateProc)GetProcAddress(dinput, "DirectInputCreateA");
	if(!proc) return -1;
	
	HRESULT hr=proc(a,b,c,d);
	if(FAILED(hr)) return hr;

	UseScrollWheel=GetPrivateProfileInt("Input", "UseScrollWheel", 1, ini)!=0;
	WheelMod=GetPrivateProfileInt("Input", "ScrollMod", 0, ini);
	LONG MouseSpeed=GetPrivateProfileInt("Input", "MouseSensitivity", 100, ini);
	if(MouseSpeed!=100) {
		AdjustMouseSpeed=true;
		MouseSpeedMod=((double)MouseSpeed)/100.0;
		MousePartX=0;
		MousePartY=0;
	} else AdjustMouseSpeed=false;

	MiddleMouseKey=GetPrivateProfileInt("Input", "MiddleMouse", 0x30, ini);
	MiddleMouseDown=false;

	if(GetPrivateProfileInt("Input", "BackgroundKeyboard", 0, ini)) BackgroundKeyboard=true; else BackgroundKeyboard=false;
	if(GetPrivateProfileInt("Input", "BackgroundMouse", 0, ini)) BackgroundMouse=true; else BackgroundMouse=false;

	*c=(IDirectInputA*)new FakeDirectInput(*c);

	return hr;
}