#ifndef WIN9X

#define DIRECTSOUND_VERSION 0x0800
#define INITGUID
#include <dsound.h>
#include <eax.h>
#include "main.h"
#include "SafeWrite.h"

#ifdef _DEBUG
#define DEBUGMESS(msg) OutputDebugStringA(msg)
#else
#define DEBUGMESS(msg)
#endif

//#define MUTESTERIO

#define UNUSEDFUNCTION { DEBUGMESS("Unused function called: " __FUNCTION__ "\r\n"); return DSERR_GENERIC; }
#define SAFERELEASE(a) { if(a) { a->Release(); a=0; } }

IDirectSound8* dsound;
IDirectSoundBuffer* DS_buffer;	//secondary buffer used to control EAX properties
LPKSPROPERTYSET EAXListener;	//IKsPropertySet for buffer (to query for EAX support)
DWORD EaxAvailable=0;
DWORD proc_addr;

DWORD _stdcall IsEAXAvailable() { return EaxAvailable; }
void _stdcall SetEAXEnvironment(DWORD env) {
    if(!EaxAvailable) return;
    if(env>25) return;
    EAXListener->Set(DSPROPSETID_EAX20_ListenerProperties, DSPROPERTY_EAXLISTENER_ENVIRONMENT, NULL, 0, &env, sizeof(DWORD));
}

static void InitEAX() {
    DSBUFFERDESC			secondary_desc;							//description structure for buffer
    LPDIRECTSOUND3DBUFFER	DS_3Dinterface;							//buffer's 3D interface

    //set waveformat info for buffer (have to do this or else CreateSoundBuffer will fail with DSERR_INVALIDPARAM)
    WAVEFORMATEX wfx;
    memset(&wfx, 0, sizeof(WAVEFORMATEX)); 
    wfx.wFormatTag = WAVE_FORMAT_PCM; 
    wfx.nChannels = 1;
    wfx.nSamplesPerSec = 44100;
    wfx.wBitsPerSample = 16;
    wfx.nBlockAlign = wfx.nChannels * wfx.wBitsPerSample / 8; 
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;

    //setup buffer
    memset (&secondary_desc, 0, sizeof(DSBUFFERDESC));
    secondary_desc.dwSize = sizeof (DSBUFFERDESC);
    secondary_desc.dwFlags = DSBCAPS_CTRL3D;									
    secondary_desc.dwBufferBytes = 1 * wfx.nAvgBytesPerSec;
    secondary_desc.lpwfxFormat = &wfx;

    dsound->CreateSoundBuffer (&secondary_desc, &DS_buffer, NULL);
    if (DS_OK != (DS_buffer->QueryInterface (IID_IDirectSound3DBuffer8, (void**)&DS_3Dinterface)))
    {
        return;
    }

    //get IKsPropertySet and check for EAX support
    DS_3Dinterface->QueryInterface (IID_IKsPropertySet, (void**)&EAXListener);
    ULONG support = 0;

    if (EAXListener->QuerySupport (DSPROPSETID_EAX20_ListenerProperties, DSPROPERTY_EAXLISTENER_ALLPARAMETERS, &support) != DS_OK)
    {
        DS_buffer->Release();
        return;
    }

    EaxAvailable=1;
}

void ResetEax() {
    SetEAXEnvironment(GetPrivateProfileIntA("Sound", "DefaultEnvironment", 1, ini));
}
static float GetPrivateProfileFloat(char* app, char* key, float def) {
	char buf[64];
	DWORD size=GetPrivateProfileStringA(app, key, "", buf, 64, ".\\eax.ini");
	if(size>=64||size==0) return def;
	float f=(float)atof(buf);
	return f;
}
static bool SetMapEnvironment2(char* mapid) {
	if(GetPrivateProfileIntA(mapid, "EnvironmentOnly", 1, ".\\eax.ini")) {
		DWORD env=GetPrivateProfileIntA(mapid, "Environment", 0, ".\\eax.ini");
		if(env<0||env>25) return false;
		SetEAXEnvironment(env);
	} else {

		long lRoom					=GetPrivateProfileIntA(mapid, "RoomLF",					0, ".\\eax.ini");
		long lRoomHF				=GetPrivateProfileIntA(mapid, "RoomHF",					0, ".\\eax.ini");
		float flRoomRolloffFactor	=GetPrivateProfileFloat(mapid, "RoomRolloffFactor",		0);
		float flDecayTime			=GetPrivateProfileFloat(mapid, "DecayTime",				0);
		float flDecayHFRatio		=GetPrivateProfileFloat(mapid, "DecayHFRatio",			0);
		long lReflections			=GetPrivateProfileIntA(mapid, "Reflections",			0, ".\\eax.ini");
		float flReflectionsDelay	=GetPrivateProfileFloat(mapid, "ReflectionsDelay",		0);
		long lReverb				=GetPrivateProfileIntA(mapid, "Reverb",					0, ".\\eax.ini");
		float flReverbDelay			=GetPrivateProfileFloat(mapid, "ReverbDelay",		    0);
		DWORD dwEnvironment			=GetPrivateProfileIntA(mapid, "Environment",			0, ".\\eax.ini");
		float flEnvironmentSize		=GetPrivateProfileFloat(mapid, "EnvironmentSize",		0);
		float flEnvironmentDiffusion=GetPrivateProfileFloat(mapid, "EnvironmentDiffusion",	0);
		float flAirAbsorptionHF		=GetPrivateProfileFloat(mapid, "AirAbsorbtionHF",		0);

		if (  dwEnvironment < 0 || dwEnvironment > 25 || flEnvironmentSize < 1.0 || flEnvironmentSize > 100.0 
			|| flEnvironmentDiffusion < 0.0 || flEnvironmentDiffusion > 1.0 || lRoom < -10000 || lRoom > 0 || lRoomHF < -10000
			|| lRoomHF > 0 || flDecayTime < 0.1 || flDecayTime > 20.0 || flDecayHFRatio < 0.0 || flDecayHFRatio > 20.0 ||
			lReflections < -10000 || lReflections > 1000 || flReflectionsDelay < 0.0 || flReflectionsDelay > 0.3 || 
			lReverb < -10000 || lReverb > 2000 || flReverbDelay < 0.0 || flReverbDelay > 0.1 || flRoomRolloffFactor < 0.0 || 
			flRoomRolloffFactor > 10.0 || flAirAbsorptionHF < -100.0 || flAirAbsorptionHF > 0.0)
		{
			return false;
		}

		EAXLISTENERPROPERTIES props = {lRoom, lRoomHF, flRoomRolloffFactor, flDecayTime, flDecayHFRatio, lReflections, 
			flReflectionsDelay, lReverb, flReverbDelay, dwEnvironment, flEnvironmentSize, 
			flEnvironmentDiffusion, flAirAbsorptionHF, 0x0000002F};	  //flags are at default values

		EAXListener->Set (DSPROPSETID_EAX20_ListenerProperties, DSPROPERTY_EAXLISTENER_ALLPARAMETERS, NULL, 0, &props, 
			sizeof (EAXLISTENERPROPERTIES));

	}

	return true;
}
void _stdcall SetMapEAXEnvironment() {
	if(!EaxAvailable) return;
	char mapid[12];
    //DWORD id=*(DWORD*)0x00519584; //This one doesn't get set on new games or loaded games...
	DWORD id=*(DWORD*)0x00631D88;
    _itoa_s(id, mapid, 10);
    if(!GetPrivateProfileIntA(mapid, "Active", 0, ".\\eax.ini")) SetMapEnvironment2("Default");
	else if(!SetMapEnvironment2(mapid)) SetMapEnvironment2("Default");
}
static void _declspec(naked) LoadMapHook() {
    _asm {
        push ecx
		mov ecx, 0x00482A68;
        call ecx;
		pop ecx;
		pushad;
        call SetMapEAXEnvironment;
        popad;
        retn;
    }
}
class FakePrimaryBuffer : public IDirectSoundBuffer
{
private:
    ULONG refs;
    IDirectSoundBuffer* real;
    IDirectSound3DListener8* listener;
    DSBUFFERDESC desc;
public:
    FakePrimaryBuffer(const DSBUFFERDESC* _desc) {
        refs=1;
        desc=*_desc;
        desc.dwFlags|=DSBCAPS_CTRL3D;
        dsound->CreateSoundBuffer(&desc, &real, 0);
        real->QueryInterface(IID_IDirectSound3DListener8, (void**)&listener);
        listener->SetPosition(0,0,0, DS3D_DEFERRED);
        listener->SetVelocity(0,0,0, DS3D_DEFERRED);
        listener->CommitDeferredSettings();
        InitEAX();
    }
    // IUnknown methods
    HRESULT _stdcall QueryInterface(REFIID a, LPVOID * b)  { UNUSEDFUNCTION; }
    ULONG _stdcall AddRef()  { return ++refs;; }
    ULONG _stdcall Release()  {
        if(!--refs) {
            SAFERELEASE(EAXListener);
            SAFERELEASE(DS_buffer);
            listener->Release();
            real->Release();
            delete this;
            return 0;
        } else return refs;
    }

    // IDirectSoundBuffer methods
    HRESULT _stdcall GetCaps(LPDSBCAPS)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetCurrentPosition(LPDWORD, LPDWORD)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetFormat(LPWAVEFORMATEX fmt, DWORD isize, LPDWORD osize)  { return real->GetFormat(fmt, isize, osize); }
    HRESULT _stdcall GetVolume(LPLONG)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetPan(LPLONG)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetFrequency(LPDWORD)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetStatus(LPDWORD)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Initialize(LPDIRECTSOUND, LPCDSBUFFERDESC)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Lock(DWORD, DWORD, void**, LPDWORD, void**, LPDWORD, DWORD)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Play(DWORD, DWORD, DWORD)  { UNUSEDFUNCTION; }
    HRESULT _stdcall SetCurrentPosition(DWORD)  { UNUSEDFUNCTION; }
    HRESULT _stdcall SetFormat(LPCWAVEFORMATEX pcfxFormat)  { return real->SetFormat(pcfxFormat); }
    HRESULT _stdcall SetVolume(LONG)  { UNUSEDFUNCTION; }
    HRESULT _stdcall SetPan(LONG)  { UNUSEDFUNCTION; }
    HRESULT _stdcall SetFrequency(DWORD)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Stop()  { UNUSEDFUNCTION; }
    HRESULT _stdcall Unlock(LPVOID, DWORD, LPVOID, DWORD)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Restore()  { UNUSEDFUNCTION; }
};

#ifdef MUTESTERIO
class NullBuffer : public IDirectSoundBuffer
{
private:
    ULONG refs;
    IDirectSoundBuffer* real;
public:
    NullBuffer(const DSBUFFERDESC* desc) {
        refs=1;
        dsound->CreateSoundBuffer(desc, &real, 0);
    }
    // IUnknown methods
    HRESULT _stdcall QueryInterface(REFIID a, LPVOID * b)  { UNUSEDFUNCTION; }
    ULONG _stdcall AddRef()  { return ++refs;; }
    ULONG _stdcall Release()  {
        if(!--refs) {
            real->Release();
            delete this;
            return 0;
        } else return refs;
    }

    // IDirectSoundBuffer methods
    HRESULT _stdcall GetCaps(LPDSBCAPS caps)  { return real->GetCaps(caps); }
    HRESULT _stdcall GetCurrentPosition(LPDWORD play, LPDWORD write)  { return real->GetCurrentPosition(play, write); }
    HRESULT _stdcall GetFormat(LPWAVEFORMATEX fmt, DWORD isize, LPDWORD osize)  { return real->GetFormat(fmt, isize, osize); }
    HRESULT _stdcall GetVolume(LPLONG plVolume)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetPan(LPLONG plPan)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetFrequency(LPDWORD pdwFrequency)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetStatus(LPDWORD pdwStatus)  { return real->GetStatus(pdwStatus); }
    HRESULT _stdcall Initialize(LPDIRECTSOUND pDirectSound, LPCDSBUFFERDESC pcDSBufferDesc)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Lock(DWORD offset, DWORD bytes, LPVOID *ptr1, LPDWORD bytes1, LPVOID *ptr2, LPDWORD bytes2, DWORD flags)  { return real->Lock(offset, bytes, ptr1, bytes1, ptr2, bytes2, flags); }
    HRESULT _stdcall Play(DWORD dwReserved1, DWORD priority, DWORD flags)  { return DS_OK; }
    HRESULT _stdcall SetCurrentPosition(DWORD dwNewPosition)  { return real->SetCurrentPosition(dwNewPosition); }
    HRESULT _stdcall SetFormat(LPCWAVEFORMATEX pcfxFormat)  { return real->SetFormat(pcfxFormat); }
    HRESULT _stdcall SetVolume(LONG lVolume)  { return DS_OK; }
    HRESULT _stdcall SetPan(LONG lPan)  { return DS_OK; }
    HRESULT _stdcall SetFrequency(DWORD dwFrequency)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Stop()  { return DS_OK; }
    HRESULT _stdcall Unlock(LPVOID ptr1, DWORD bytes1, LPVOID ptr2, DWORD bytes2)  { return real->Unlock(ptr1, bytes1, ptr2, bytes2); }
    HRESULT _stdcall Restore()  { UNUSEDFUNCTION; }
};
#endif

class FakeMonoBuffer : public IDirectSoundBuffer
{
private:
    ULONG refs;
    IDirectSoundBuffer* real;
    IDirectSound3DBuffer8* emitter;
    DSBUFFERDESC desc;
    WAVEFORMATEX fmt;
public:
    FakeMonoBuffer(const DSBUFFERDESC* _desc) {
        refs=1;
        desc=*_desc;
        fmt=*desc.lpwfxFormat;
        desc.lpwfxFormat=&fmt;
        desc.dwFlags|=DSBCAPS_CTRL3D;
        dsound->CreateSoundBuffer(&desc, &real, 0);
        real->QueryInterface(IID_IDirectSound3DBuffer8, (void**)&emitter);
        emitter->SetPosition(0, 1, 0, DS3D_IMMEDIATE);
    }
    // IUnknown methods
    HRESULT _stdcall QueryInterface(REFIID a, LPVOID * b)  { UNUSEDFUNCTION; }
    ULONG _stdcall AddRef()  { return ++refs;; }
    ULONG _stdcall Release()  {
        if(!--refs) {
            real->Release();
            emitter->Release();
            delete this;
            return 0;
        } else return refs;
    }

    // IDirectSoundBuffer methods
    HRESULT _stdcall GetCaps(LPDSBCAPS caps)  { return real->GetCaps(caps); }
    HRESULT _stdcall GetCurrentPosition(LPDWORD play, LPDWORD write)  { return real->GetCurrentPosition(play, write); }
    HRESULT _stdcall GetFormat(LPWAVEFORMATEX fmt, DWORD isize, LPDWORD osize)  { return real->GetFormat(fmt, isize, osize); }
    HRESULT _stdcall GetVolume(LPLONG plVolume)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetPan(LPLONG plPan)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetFrequency(LPDWORD pdwFrequency)  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetStatus(LPDWORD pdwStatus)  { return real->GetStatus(pdwStatus); }
    HRESULT _stdcall Initialize(LPDIRECTSOUND pDirectSound, LPCDSBUFFERDESC pcDSBufferDesc)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Lock(DWORD offset, DWORD bytes, LPVOID *ptr1, LPDWORD bytes1, LPVOID *ptr2, LPDWORD bytes2, DWORD flags)  { return real->Lock(offset, bytes, ptr1, bytes1, ptr2, bytes2, flags); }
    HRESULT _stdcall Play(DWORD dwReserved1, DWORD priority, DWORD flags)  { return real->Play(dwReserved1, priority, flags); }
    HRESULT _stdcall SetCurrentPosition(DWORD dwNewPosition)  { return real->SetCurrentPosition(dwNewPosition); }
    HRESULT _stdcall SetFormat(LPCWAVEFORMATEX pcfxFormat)  { return real->SetFormat(pcfxFormat); }
    HRESULT _stdcall SetVolume(LONG lVolume)  { return real->SetVolume(lVolume); }
    HRESULT _stdcall SetPan(LONG lPan)  { UNUSEDFUNCTION; }
    HRESULT _stdcall SetFrequency(DWORD dwFrequency)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Stop()  { return real->Stop(); }
    HRESULT _stdcall Unlock(LPVOID ptr1, DWORD bytes1, LPVOID ptr2, DWORD bytes2)  { return real->Unlock(ptr1, bytes1, ptr2, bytes2); }
    HRESULT _stdcall Restore()  { UNUSEDFUNCTION; }
};

class FakeDirectSound : public IDirectSound
{
private:
    ULONG refs;
public:
    FakeDirectSound() {
        refs=1;
    }
    // IUnknown methods
    HRESULT _stdcall QueryInterface(REFIID a, LPVOID * b)  { UNUSEDFUNCTION; }
    ULONG _stdcall AddRef()  { return ++refs;; }
    ULONG _stdcall Release()  {
        if(!--refs) {
            dsound->Release();
            delete this;
            return 0;
        } else return refs;
    }

    // IDirectSound methods
    HRESULT _stdcall CreateSoundBuffer(LPCDSBUFFERDESC desc, LPDIRECTSOUNDBUFFER *buffer, LPUNKNOWN)  {
        if(!desc->lpwfxFormat) *buffer=new FakePrimaryBuffer(desc);
        else if(desc->lpwfxFormat->nChannels==1) *buffer=new FakeMonoBuffer(desc);
        else if(desc->lpwfxFormat->nChannels==2) {
#ifdef MUTESTERIO
			*buffer=new NullBuffer(desc);
#else
            dsound->CreateSoundBuffer(desc, buffer, 0);
#endif
        }
        else UNUSEDFUNCTION;
        return DS_OK;
    }
    HRESULT _stdcall GetCaps(LPDSCAPS pDSCaps)  { return dsound->GetCaps(pDSCaps); }
    HRESULT _stdcall DuplicateSoundBuffer(LPDIRECTSOUNDBUFFER pDSBufferOriginal, LPDIRECTSOUNDBUFFER *ppDSBufferDuplicate)  { UNUSEDFUNCTION; }
    HRESULT _stdcall SetCooperativeLevel(HWND hwnd, DWORD dwLevel)  { return dsound->SetCooperativeLevel(hwnd, dwLevel); }
    HRESULT _stdcall Compact()  { UNUSEDFUNCTION; }
    HRESULT _stdcall GetSpeakerConfig(LPDWORD pdwSpeakerConfig)  { UNUSEDFUNCTION; }
    HRESULT _stdcall SetSpeakerConfig(DWORD dwSpeakerConfig)  { UNUSEDFUNCTION; }
    HRESULT _stdcall Initialize(LPCGUID pcGuidDevice)  { UNUSEDFUNCTION; }
};

HRESULT _stdcall FakeDSoundCreate(DWORD,IDirectSound** iface,DWORD) {
    DirectSoundCreate8(0, &dsound, 0);
    *iface=new FakeDirectSound();

    //This hook doesn't get hit on save game loads or new games...
    SafeWrite32(0x00482B5E, (DWORD)&LoadMapHook - 0x00482B62);

    //This hook does, but no way to get the current map id.
    //SafeWrite32(0x004831AE, (DWORD)&LoadMapHook - 0x004831B2);

    return DS_OK;
}


#endif